/*-- Copyright (c) 2011 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved. --*/ 
/*
Macro         PAM 5.1 PAM_COMPARE.SAS
Purpose       Compares a PAM Mart table with and (_RK or _SK) to an incoming warehouse table of the same format 
                with the exception of the _RK/_SK field, process_dttm, record_status_flg, 
                load_dttm and current_num.  The Valid_to_dttm field is expected to be populated either from the warehouse or
                use of the pam_eedate.sas macro execution. Its field name is obtained from the data dictionary but must include
                valid_to_dttm in its name.A check will be made to see if the provided valid_to_dttm field is correct - meaning that 
                it does not occur before the valid_from_dttm 
              This code also checks for columns to be added or removed from the final output.
              Columns are only removed when the compare output table is empty else no columns are removed
              Columns can be added as new columns are included by the site
              Note:  PROCESS_DTTM, RECORD_STATUS_FLG, LOAD_DTTM  are unique to the compare or source table and are not removed
              Updated 5.2 - cleaned up current_num and removed hard coding for valid_to_dttm as this can be any named valid_to_dttm field
              
                 
Macro Name    pam_compare
DBMS          SAS
Created       10/14/2011 
Copyright (c) 2011 by SAS Institute Inc., Cary, NC 27513
*/

/* Parameters 
    SOURCEDS   - Required - The source dataset name - required parameter and the table must exist 
                          - from the incoming data (typically from the warehouse) 
    SOURCELIB  - The source library - defaults to WORK if none provided 
    COMPAREDS  - Required  - Compare dataset output table (typically the data mart table) 
    COMPARELIB - The libref for the compare datasets.  Defaults to the macro for the PAM MART.
    PRIMARYKEY - Required - The primary key of the output table. The _RK or _SK field
    BUS_PRIM_KEY - Required - the business primary key of the output table.  The key does not include the valid_to_dttm; 
    DROPCOLS   -  Columns not in common between the two tables.  This is used to drop these columns for the digest value
                 only.  Where appropriate these fields are maintained in the output table.
    IGNORE_INPUT_COLUMNS - list of columns from the input table to not include in the output table.
*/ 
                     
%macro pam_compare(sourceds=, sourcelib=WORK, compareds= , comparelib=&pam_dmlibref, primarykey=, bus_prim_key=, 
                   dropcols=PROCESS_DTTM RECORD_STATUS_FLG LOAD_DTTM,
                   ignore_input_columns=);                                         
options &mprint &source &notes;

%global debug;

/* last_load_dttm used for load_status table for new records  and from_dttm used in identifying the valid_from_dttm field used in business keys */ 
%global last_load_dttm pam_tmpRC pam_compare_recnt; 
%let last_load_dttm= ; 
%let pam_tmpRC = 0;
%let to_dttm = ; /* this used for check for current_num record */ 

/* ---------------------------------------------------- */
/* Parameter Validation                                 */
/* ---------------------------------------------------- */
%LET SOURCEDS  =%nrbquote(&SOURCEDS);
%LET SOURCELIB =%nrbquote(&SOURCELIB);
%LET COMPAREDS =%nrbquote(&COMPAREDS);
%LET COMPARELIB=%nrbquote(&COMPARELIB);
%LET PRIMARYKEY=%nrbquote(&PRIMARYKEY);
%LET BUS_PRIM_KEY=%nrbquote(&BUS_PRIM_KEY);
%LET DROPCOLS  =%nrbquote(&DROPCOLS);

%let dropcols        =%kupcase(&dropcols);                                                                                                    
%let sourceds        =%kupcase(&sourceds); 
%let sourcelib       =%kupcase(&sourcelib); 
%let compareds       =%kupcase(&compareds); 
%let comparelib      =%kupcase(&comparelib);
%let primarykey      =%kupcase(&primarykey);                                                                                                  
%let bus_prim_key    =%kupcase(&bus_prim_key);  
%let pam_dmlibref    =%kupcase(&pam_dmlibref);

 
/* sourceds is the new source table coming in.  Compareds is the final table used to determine new, missing, changed */ 
%if &sourceds = %then %do;
    %put ERROR:  The sourceds must be provided;
    %let pam_tmpRC = 1;
%end;
%if &sourcelib = %then %do;
     %put ERROR:  The source library must be provided;
     %let pam_tmpRC = 1;
%end;
%if &compareds = %then %do;
     %put ERROR:  The compare dataset must be provided;
     %let pam_tmpRC = 1;
%end;
%if &comparelib = %then %do;
     %put ERROR:  The compare library must be provided;
     %let pam_tmpRC = 1;
%end;
%if &primarykey = %then %do;
     %put ERROR:  The primary key must be provided;
     %let pam_tmpRC = 1;
%end;
%if &bus_prim_key = %then %do;
     %put ERROR:  The business primary key must be provided;
     %let pam_tmpRC = 1;
%end;


/* ---------------------------------------------------- */
/* End Parameter Validation                             */
/* ---------------------------------------------------- */
/* Further checks are below for the compare and source table existance */ 
 
%if &pam_tmpRC = 0 %then %do; /* check for parameters */ 
/* ensure work files are empty */   
proc datasets lib = work nolist nowarn memtype = (data view);                                                                                                                                                                   
   delete missing_records nochange pam_compare_digest dict temp pam_master_compare_digest new_target_lookup_tbl                                                                                                                 
          new newck change missing_records newfinal newtest target_lookup_tbl compare_col_list ;                                                                                                                                                 
quit;

/* ---------------------------------------------------------------- */
/* Macro checking for number of observations in table - used below  */
/* Embedded macro pam_compare_record_check                          */
/* ---------------------------------------------------------------- */
%macro pam_compare_record_check(input_file=&infileref); 
  /* the input_tbl and input_lib are coming in as macros so add the extra . */ 
  options &mprint &source &notes;  
   %let pam_recCheckExist = %eval(%sysfunc(exist(&input_file, DATA)) or 
         %sysfunc(exist(&input_file, VIEW))); 
   %if (&pam_recCheckExist) %then
   %do;
           
      proc contents data = &input_file out = work.pam_contents(keep = nobs) noprint; 
      run; 
      
      data _null_; 
         set work.pam_contents (obs = 1); 
         call symputx("pam_compare_recnt", nobs); 
      run;
      /* write out count of records in compare */ 
      %put PAM_COMPARE_RECNT=&PAM_COMPARE_RECNT;     
      proc datasets lib = work nolist nowarn memtype = (data view);
         delete pam_contents;
      quit;
   %end;
%mend pam_compare_record_check;
 

/* put the DROPCOLS into their own individual macro variables */
%local nwords;
%let nwords=1;
%do %until(&word= );
   %let word=%kscan(&dropcols,&nwords,%str( )=);
   %if &word ne  %then %do;
     %local drop&nwords;
     %let drop&nwords=&word;
     %let nwords=%eval(&nwords+1);
   %end;
%end;
%let nwords=%eval(&nwords-1);


/* put the business primary key columns into their own individual macro variables to check for needed commas */
/* used to subset the dictionary */ 
%local buswords;
%let buswords=1;
%let check_date = 0; 
%let pam_fromVar = ;
%do %until(&word= );
  %let word=%kscan(&bus_prim_key,&buswords,%str( )=);
  %let check_date = %kindex(%kupcase(&word),%str(VALID_FROM_DTTM)); 

  %if (&check_date gt 0) %then 
     %let pam_fromVar = &word;

  %if &word ne  %then %do;
    %local bkp&buswords bkp_minus_from&buswords;
    %let bkp&buswords=&word;
    %let buswords=%eval(&buswords+1);
    %if &check_date eq 0 %then %do;
      %let bkp_minus_from&buswords=&word;
    %end;
  %end;
%end;
%let buswords=%eval(&buswords-1);

/* Add commas and quotes to for multiple business key field resolution for hash code */ 
/* Else add quote around prim_bus_key as it is removed in the hash statement */ 
%if %eval(&buswords) > 1 %then %do;                                                                                                 
        data work.temp;                                                                                                                  
                length name $32;                                                                                                               
                %do i=1 %to &buswords;                                                                                                         
                    name = "&&bkp&i";                                                                                                          
                    output;                                                                                                                    
                %end;                                                                                                                          
                run;                                                                                                                           
                                                                                                                                               
                data _null_;                                                                                                                   
                   length pamfld  busfld  $1000;                                                                                                        
                    set work.temp end=eof;                                                                                                      
                   retain pamfld busfld;                                                                                                              
                    if _n_=1 then do;                                                                                                          
                           pamfld = quote(kcompress(name)); 
                           busfld = kcompress(name); 
                    end;                                                                                                                       
                    else do;                                                                                                                   
                       pamfld = kcompress(pamfld)||' , '||quote(kcompress(name));  
                       busfld = kcompress(busfld)||' , '||kcompress(name);   
                    end;                                                                                                                       
                   if eof then do;                                                                                                             
                      call symputx('PRIM_FLD',kcompress(pamfld)); 
                      call symputx('BUS_FLD',kcompress(busfld)); 
                    end;                                                                                                                       
                 run;       
%end; 
%else %do; 
    %let PRIM_FLD="&bus_prim_key"; 
    %let BUS_FLD = &bus_prim_key; 
%end; 

/* put the PRIMARY Key into their own individual macro variables */
%local pkwords;
%let pkwords=1;
%do %until(&word= );
   %let word=%kscan(&primarykey,&pkwords,%str( )=);
   %if &word ne  %then %do;
     %local drop&pkwords;
     %let pk&pkwords=&word;
     %let pkwords=%eval(&pkwords+1);
   %end;
%end;
%let pkwords=%eval(&pkwords-1);
%let RKSKVAL = ; 

/* ensure work temp file is empty */   
proc datasets lib = work nolist nowarn memtype = (data view);                                                                                                                                                                   
   delete temp ;                                                                                                                                                 
quit;

/* Add commas and quotes to for multiple primary key field resolution for hash code */ 
/* Else add quote around primary key as it is removed in the hash statement */ 
 %if %eval(&buswords) > 1 %then %do;                                                                                                 
        data work.temp;                                                                                                                  
                length name $32;                                                                                                               
                %do i=1 %to &pkwords;                                                                                                         
                    name = "&&pk&i";                                                                                                          
                    output;                                                                                                                    
                %end;                                                                                                                          
                run;                                                                                                                           
                                                                                                                                               
                data _null_;                                                                                                                   
                   length prifld $1000;                                                                                                        
                    set work.temp end=eof;                                                                                                      
                    retain prifld;                                                                                                              
                    if _n_=1 then do;                                                                                                          
                       prifld = quote(kcompress(name));                                                                                     
                    end;                                                                                                                       
                    else do;                                                                                                                   
                       prifld = kcompress(prifld)||','||quote(kcompress(name));                                                                  
                    end; 
                    pos  = kindex(kupcase(name),'_RK');
                    pos1 = kindex(kupcase(name),'_SK'); 
                    if pos > 1  or pos1 > 1 then do;
                        call symputx('RKSKVAL',kcompress(kupcase(name)));
                    end; 
                    
                  put prifld=;                                                                                                                 
                   if eof then do;                                                                                                             
                      call symputx('PRIMARYKEY_FLD',kcompress(prifld));                                                                               
                    end;                                                                                                                       
                 run;       
%end; 
%else %let PRIMARYKEY_FLD="&primarykey"; 

%put RKSKVAL=&rkskval; 

%if "&RKSKVAL"  =  '' %then %do;
  %put ERROR: Compare needs a target comparison table with an _RK or _SK value.  Process aborted ;
  %let pam_tmpRC = 1; 
%end;
%Else %do; /* Finish the rest of the logic _RK or _SK value found  */ 

/* compareds needs the libref while the dictionary needs to be work */                                                                     
%let ck=%kindex(%kupcase(%superq(compareds)),%kupcase(%quote(PAMMART)));                                                                      
%if &ck < 1 %then %let ck=%kindex(%kupcase(%superq(compareds)),%kupcase(%quote(&pam_dmlibref)));                                              
%if &ck lt 1 %then %do;                                                                                                                 
   %let comparenolib   =%kupcase(&compareds);                                                                                                  
   %let compareds      =%kupcase(&pam_dmlibref..&compareds);                                                                                   
%end;                                                                                                                                   
%else %do;                                                                                                                              
  %let ck=%kindex(%kupcase(%superq(compareds)),.);                                                                                           
  %if &ck ge 1 %then %do;                                                                                                               
      %let comparenolib=%kupcase(%ksubstr(&compareds,&ck+1));                                                                                 
  %end;                                                                                                                                 
%end;

/* Housekeeping on the saved output pametllib data */ 
proc datasets lib=&pam_etlibref nolist nowarn memtype = (data view);
   delete &comparenolib._change &comparenolib._new &comparenolib._missing; 
quit;

/* Compare obs check */ 
/* Check if the compare table is populated or empty - if empty then all are new and sync up columns from source */ 
%let pam_recCheckExist=0;
%let pam_compare_recnt=0; 
%let infileref=&comparelib..&comparenolib;

/* Check for records in the compare table - for removal or addition of other columns */ 
%pam_compare_record_check(input_file=&infileref);

%if %eval(&pam_compare_recnt) gt 0 %then %let pam_compare_comp = 1;
%else %let pam_compare_comp = 0; 

/* Source  */ 
/* sourceds can be a work or permanent libref while the dictionary needs to be work */ 
%let ck=%kindex(%kupcase(%superq(sourceds)),'.');
%if &ck ge 1 %then %do;
    %if "&sourcelib" ne "" %then %do;
       %let sourcelib = %kupcase(%ksubstr(%superq(sourceds)),1,&ck-1);
    %end;
    %let sourcenolib=%kupcase(%ksubstr(%superq(sourceds)),&ck+1);
%end;
%else %do;
    %let sourcenolib=%kupcase(&sourceds);
%end; 
%let sourceds=&sourcelib..&sourcenolib;
                   
/* Record Check on Source Target Table */ 
%global pam_compare_recnt; 
%let pam_recCheckExist=0;
%let pam_compare_recnt=0; 
%let infileref=&sourcelib..&sourcenolib; 

/* Check for records in the source table */ 
%pam_compare_record_check(input_file=&infileref);

%if %eval(&pam_recCheckExist) gt 0 %then %let pam_compare_source = 1;
                                   %else %let pam_compare_source = 0; 
                                   
/* if no source records found then this comparison stops - goes to bottom */ 
%if &pam_compare_source = 0 %then %do;
   %put ERROR:  No source table found for comparison;  
   %put ERROR:  Check that the options input source table and library match what is included in the options;  
   %put ERROR:  No records added to the target table;
   %let pam_tmpRC = 1; 
%end; 

/* Check for source records for comparison - if no source records exit with no changes or additions  */ 
%If &pam_compare_source ne 0 %then %do; 

/* Remove tables used in this comparison */ 
proc datasets lib = work nolist nowarn memtype = (data view);                                                                           
   delete missing_records nochange pam_compare_digest dict dict_source dict_source_all;                                                                                                         
quit;

/* check the columns in ignore_input_columns to  */
/* see if each one exists, and if there are      */
/* any drop columns, then remove them from the   */
/* input table, and save the table to            */
/* work.sourcenolib and reset the sourceds and   */
/* sourcelib macro vars                          */
%if (&ignore_input_columns ne ) %then 
%do;
   data _null_;
      dsid = open("&sourceds");
      if (dsid gt 0);
      i = 0;
      length col $100;
      length dropList $1000;
      do until (col eq "");
         i + 1;
         col = scan("&ignore_input_columns", i, " ");
         if (col ne "") then 
         do; 
            if (varnum(dsid, col) gt 0) then 
               dropList = catx(" ", dropList, col);
         end;
      end;
      dsid = close(dsid);
      if (dropList ne "") then 
      do; 
         call execute("data work.&sourcenolib;");
         call execute("   set &sourceds;");
         call execute("   drop " !! strip(dropList) !! ";");
         call execute("run;");
         call symput("sourceds", "WORK.&sourcenolib");
         call symput("sourcelib", "WORK");
      end;
   run;
%end;  /* if - ignore_input_columns is not blank */
 
/* Check for the physical columns in the source table as they can contain hierarchies not in the SAS metadata mapping */                  
/*  of the table and mapping in the job at this time */                                                                                 
  proc sql;                                                                                                                             
   create table dict_source_all                                                                                                                    
   as select name, type, length, label from dictionary.columns                                                                                               
   where kupcase(libname)="&sourcelib" and kupcase(memname) = "&sourcenolib" ;
   
  quit; 
  
  
/* Drop columns not for comparison - ones only in specific tables */ 
/* Ouput the length of the business key column.  needed below */ 
 
  data dict(keep=name);                                                                                                                            
     set dict_source_all end=eof;                                                                                                                  
     where kupcase(name) not in ("&primarykey") ;
     %if &buswords ne '' %then %do;
       %do i=1 %to &buswords;
           if kupcase(name) = %kupcase("&&bkp&i") then do;
               call symputx("bkp&i._length",length,'L');
               if kupcase(type) = 'CHAR' then do;
                 call symputx("bkp&i._type",'$','L'); 
               end;
               else do;
                 call symputx("bkp&i._type",.,'L'); 
               end; 
               delete;
            end; 
        %end;
     %end;     
     %if &nwords ne '' %then %do;
         %do i=1 %to &nwords;
            if kupcase(name) = %kupcase("&&drop&i") then delete;  
         %end;
     %end;
     /* Do not include the current_num field in the comparison - that will be updated at the end */ 
     if kupcase(name) = 'CURRENT_NUM' then delete;
     if kindex(kupcase(name),'VALID_TO_DTTM') gt 0 then do;
        call symputx('TO_DTTM',name);
     end; 
  run;
  
  /* Check if the valid_to_dttm field is included.  If excluded there is a problem and the process stops */ 
  %if &to_dttm = %then %do;
       %put ERROR:  The valid_to_dttm field is missing.  It is included from warehouse to mart 
                    or generated by pam_eedate transformation in an earlier step and not deleted;
       %let pam_tmpRC = 1;
  %end;
  
%do i=1 %to &buswords;
   %put bkp&i._length = &&bkp&i._length;
   %put bkp&i._type=&&bkp&i._type; 
%end; 
    /* arrange the source by sorted name order for comparison */ 
  proc sort data=dict;
   by name;
  run; 
/* keep dictionary of source columns to check with what is in the compare - for additions */ 
  data dict(drop=sourcevals sourcevals_commas);
    set dict end=eof;
    by name;
     attrib sourcevals sourcevals_commas length=$1000;                                                                                                      
     retain sourcevals sourcevals_commas;
     if _n_=1 then do;
       sourcevals=COMPBL(name); 
       sourcevals_commas=COMPBL(name);
     end; 
     else do; 
        sourcevals        = COMPBL(sourcevals)||COMPBL(name); 
        sourcevals_commas = COMPBL(sourcevals_commas)||','||COMPBL(name);
     end;  
     
     if eof then do;                                                                                                                    
       call symputx('sourceVALS',sourcevals,'F'); 
       call symputx('sourceVALS_COMMAS',sourcevals_commas,'F'); 
     end;                                                                                                                               
   run; 
   
%let sourcevals        = %kupcase(&sourcevals); 
%let sourcevals_commas = %kupcase(&sourcevals_commas); 

%put sourcevals=&sourcevals;
%put sourcevals_commas=&sourcevals_commas;

 proc datasets lib=work nolist nowarn memtype = (data view); ;
  delete dict compare_col_list add_to_compare remove_from_compare dict;
 quit;

/* Check for the physical columns in the compare table as they can contain hierarchies not in the SAS metadata mapping */                  
/*  of the table and mapping in the job at this time */

%let comparevals=; 
%let comparevals_commas=; 

  proc sql;                                                                                                                             
   create table dict                                                                                                                    
   as select name from dictionary.columns                                                                                               
   where kupcase(libname)="&comparelib" and kupcase(memname) = "&comparenolib" ;                                                         
  quit; 
  
/* Check what columns are added for compare table output.  Contains the name type length label format and informat */ 
  /* uppercase source - as all should be uppercase */ 
  data work.dict_source_all;
     set work.dict_source_all;
     name=kupcase(name);
  run; 
  proc sort data=dict_source_all;
    by name;
  run; 
 
  data work.compare_col_list;
    set dict;
    name=kupcase(name);
  run;
  /* needed for declare listing for new records section for append */ 
  proc sort data=compare_col_list;
    by name;
  run; 
 /* ------------------------------------------------------------------------------------------------------------- */
 /* If compare is populated add in new source columns and keep all compare columns                                */ 
 /* Does not include the primary or business keys - uses the prim_fld and primarykey_fld comma quoted values for  */
 /* logic.  Add and remove tables will be used in the add section below                                           */ 
 /* NOTE:  PROCESS_DTTM, RECORD_STATUS_FLG, LOAD_DTTM are excluded from the add to compare as these are only in   */
 /*        one table not both                                                                                     */ 
 /* ------------------------------------------------------------------------------------------------------------- */
  
  data compare_col_list(keep=name)
       add_to_compare 
       remove_from_compare(keep=name);
    merge compare_col_list(in=a) dict_source_all(in=b);
    by name;
         if a and b then output compare_col_list;
    else if b and not a then do;
        output compare_col_list;
        if kupcase(name) not in ('PROCESS_DTTM','RECORD_STATUS_FLG','LOAD_DTTM') then 
        output add_to_compare; 
    end; 
    else if a and not b then do;
      if name not in (&prim_fld) and name not in (&primarykey_fld) and 
         name not in ('PROCESS_DTTM','RECORD_STATUS_FLG','LOAD_DTTM','CURRENT_NUM') then    
        output remove_from_compare;
    end; 
    else delete;
  run;
  
  /* Set up for additional columns to add to compare table */ 
  %let ADD_TO_COMPARE = 0; 
  /* check for obs in the add to compare */ 
  %pam_compare_record_check(input_file=add_to_compare);
  %if %eval(&pam_compare_recnt) gt 0 %then %do;
    %let ADD_TO_COMPARE = 1; 
    data _null_;
          set add_to_compare end=eof;
          length newtype $25 fldname $1000;
          i=_n_;
            if kupcase(type) = 'CHAR' then newtype=cats('VARCHAR(', length, ')');
            else if kupcase(type) = 'NUM' then newtype = cats('NUMERIC(', length, ')'); 
            fldname = 'ADD '||kcompress(NAME)||' '||kcompress(newtype)||' '||
                  "LABEL='"||ktrim(kleft(LABEL))||"' ";
            hold_name = kcompress('ADD_NAME'||i);
            put "NOTE:  Field " NAME "added with the following attributes " fldname; 
            call symputx(hold_name,fldname,'F'); 
                
          if eof then do;
            call symputx('ADD_TO_COMPARE',i); 
          end; 
    run;
  %end; /* end add to compare section */ 
  
  /* set up for any columns to be removed from initial compare - nothing dropped if tables are already populated */ 
  /* No business keys or primary keys will be removed */ 
  %let REMOVE_FROM_COMPARE = 0; 
  %if &pam_compare_comp = 0 %then %do;
    /* check for obs in the remove_from_compare table */ 
      %pam_compare_record_check(input_file=remove_from_compare); 
      %if %eval(&pam_compare_recnt) gt 0 %then %do; 
         %let REMOVE_FROM_COMPARE = 1; 
         data _null_;
          set remove_from_compare end=eof;
          length fldname $1000 dropnm $32;
          i=_n_;
          fldname = "DROP COLUMN "||kcompress(NAME); 
          dropnm=cats('DROP_NAME', i); 
          call symputx(dropnm,fldname,'F');
          if eof then do;
            call symputx('REMOVE_FROM_COMPARE',i,'F');
          end;
        run; 
      %end; /* remove_from_compare section */ 
  %End; /* pam_compare_comp equal 0  */  

 /* remove any duplicates */ 
 proc sort data=compare_col_list nodupkey;
   by name;
 quit; 
 
 /* Delete the dictionary work table */ 
 proc datasets lib=work nolist nowarn memtype = (data view); ;
  delete dict_source;
 quit;
  
  data _null_;
    length comp_col_list $1000; 
    set compare_col_list end=eof;
    by name;
    retain comp_col_list;
    if _n_ = 1 then do;
      comp_col_list = kcompress(name);
    end;
    else do;
      comp_col_list = kleft(ktrim(comp_col_list))||' '||kcompress(name);
    end;
    if eof then do;
      call symput('COMP_COL_LIST',kleft(ktrim(comp_col_list))); 
    end;
  run; 
  
  data dict(keep=name);                                                                                                                            
     set dict end=eof;                                                                                                                  
     where kupcase(name) not in ("&primarykey"); 
     /* subset out the primary business key */ 
     %if &buswords ne '' %then %do;
            %do i=1 %to &buswords; 
                if kupcase(name) = %kupcase("&&bkp&i") then delete;
            %end;
     %end; 
     /* remove the drop variables */ 
     %if &nwords ne '' %then %do;
         %do i=1 %to &nwords;
              if kupcase(name) = %kupcase("&&drop&i") then delete;  
         %end;
     %end;
     /* Do not include the current_num field in the comparison - that will be updated at the end */ 
     if kupcase(name) = 'CURRENT_NUM' then delete; 
   run; 
   
  /* arrange the compare by sorted name order for comparison */ 
  proc sort data=dict;
   by name;
  run; 
  
  %put Compare Dictionary fields ;
  %let TARGETCNT = 0; 
  data dict(drop=comparevals comparevals_commas count);  
    set dict end=eof;
     by name;
     attrib comparevals  comparevals_commas length=$1000;  
     length count 8; 
     retain comparevals comparevals_commas count;
     count + 1; 
     if _n_=1 then do;
       comparevals       =COMPBL(name); 
       comparevals_commas=COMPBL(name);
     end; 
     else do; 
        comparevals        = COMPBL(comparevals)||COMPBL(name); 
        comparevals_commas = COMPBL(comparevals_commas)||','||COMPBL(name);
     end; 
     call symputx(cats('TARGET_FIELD', count),name); 
     if eof  and count ge 1 then do;
          call symputx('COMPAREVALS',comparevals,'F'); 
          call symputx('COMPAREVALS_COMMAS',comparevals_commas,'F');
          call symputx('TARGETCNT',count);
     end; 
run; 

%do i=1 %to &targetcnt;
  %put target_field&i= &&target_field&i;
%end; 
%let comparevals        = %kupcase(&comparevals); 
%let comparevals_commas = %kupcase(&comparevals_commas); 

/* printing out comparevals and sourcevals used for comparison */ 
%put sourcevals_commas = &sourcevals_commas;

%put comparevals_commas = &comparevals_commas;

%put prim_fld = &prim_fld;

%let ETL_CHANGE         = %kupcase(&pam_etlibref..&sourcenolib._change);
%let ETL_NEW            = %kupcase(&pam_etlibref..&sourcenolib._new);
%let ETL_MISSING        = %kupcase(&pam_etlibref..&sourcenolib._missing); 

/* ---------------------------------------------------------------------------------------- */
/* Start processing records from comparison setup above                                     */
/* sort the source by the primary business key for comparison                               */ 
/* ---------------------------------------------------------------------------------------- */ 
%let sourcelibck = &sourcelib..&sourcenolib;
 
proc sort data=&sourcelibck out=work.sourcenolib;
  by &bus_prim_key ;  
run; 

%put bus_prim_key = &bus_prim_key;
%put bkp_buswords= &&bkp&buswords;
%put bkp_minus_from&buswords = &&bkp_minus_from&buswords; 


data work.sourcenolib;
   set work.sourcenolib;
   by &bus_prim_key ;  
   attrib CURRENT_NUM length=8 ;
run; 

proc sort data=work.sourcenolib;
  by &bus_prim_key;
run; 

/* create compare digest from the master data table and hold onto it until change section */  
/* create from the read to the target (compare) table a listing of the primary and first business key for lookup */ 
data work.pam_master_compare_digest(keep= &primarykey &bus_prim_key compare_digest buskey_digest)
     work.target_lookup_tbl(keep=&primarykey &bkp1); 
  length compare_digest  buskey_digest $32;
  set &compareds (keep=&primarykey &bus_prim_key &comparevals); 
  compare_digest=put(md5(catq(' ',&comparevals_commas)), $hex32.); 
  buskey_digest =put(md5(catq(' ',&bus_fld)), $hex32.);
  output work.pam_master_compare_digest;
  output work.target_lookup_tbl;
run; 

   
/* source/compare lookup: hash lookup method  */   
/* Prim_fld is the breakout with double quotes and commas of the primary business field */ 
/* these are expected to be the same on both tables */ 

data WORK.CHANGE(drop=compare_digest source_digest)
     WORK.NEW
     WORK.NOCHANGE
      ; 
     
     length source_digest $32; 
    /* drop source_digest compare_digest rc; */ 
                                                                                                                                                                                                            
     if 0 then                                                                                                                            
       set work.pam_master_compare_digest;                                                                                                     
                                                                                                                                        
     if _N_ eq 1 then                                                                                                                     
     do;                                                                                                                                  
        declare hash hct(dataset: 'WORK.PAM_MASTER_COMPARE_DIGEST', hashexp: 10);                                                               
                     hct.defineKey(&prim_fld);                                                                                         
                     hct.defineData(&prim_fld, "compare_digest");                                                                      
                     hct.defineDone(); 
     end;                                                                                                                                 
                                                                                                                                        
   set work.sourcenolib end=EOF;                                                                                                         
                                                                                                                                        
   /* create digest  */                                                                                                                 
   source_digest = put(md5(catq(' ',&sourcevals_commas)), $hex32.);
   buskey_digest = put(md5(catq(' ',&bus_fld)), $hex32.);
                                                                                                                                         
   /* source/compare match  */                                                                                                          
   if hct.find() eq 0 then                                                                                                              
   do;
      /* source/compare match: no change detected  */                                                                                   
      if source_digest eq compare_digest then                                                                                           
         output WORK.NOCHANGE; 
                                                                                                                                       
      /* source/compare match: change detected  */                                                                                      
      else if source_digest ne compare_digest then                                                                                      
         output WORK.CHANGE ; 
                                                                                                                                              
      /* remove found key + data  */                                                                                                    
      hct.remove();                                                                                                                     
   end;                                                                                                                                 
   /* source table: new records  */                                                                                                     
   else if hct.find() ne 0 then do; 
      output WORK.NEW;                                                                                                     
   end;
   
   if EOF then do;                                                                                                                         
      rc = hct.output(dataset: 'WORK.MISSING_RECORDS');
   end;                                                           
run;  

/* drop the compare digest  */                                                                                                          
data &ETL_MISSING;                                                                                                         
   set work.missing_records(drop = compare_digest);                                                                                
run; 

/* make hard copy of change and use work below for less io */ 
data &ETL_CHANGE;
    set work.change;
run; 
data &ETL_NEW;
    set work.new;
run; 

/* ------------------------------------------------------------------------------------------- */  
/* --- Modify the compare table to add columns.  Only remove columns if the table is empty -- */
/* --- No business keys are removed                                                         -- */ 
/* ------------------------------------------------------------------------------------------- */ 
  /* Handle the new columns and removed columns from the compare data when table is empty */ 
  %if &add_to_compare gt 0 or &remove_from_compare gt 0 %then %do; 
   proc sql;
     alter table &comparelib..&comparenolib
     %if &add_to_compare gt 0 %then %do; 
       %do i=1 %to &add_to_compare;
          &&add_name&i
       %end;
     %end; 
     %if &remove_from_compare gt 0 %then %do; 
       %do i=1 %to &remove_from_compare;
          &&drop_name&i
       %end;
     %end;
     ;;
   quit; 
  %end; /* handle new columns and removed columns section */ 

/* ------------------------------------------------------------------------------------------- */  
/* --- Perform the Change Record Update to the Target Table (compareds)                    --- */
/* ------------------------------------------------------------------------------------------- */ 
  
 /* Record Check on WORK.CHANGE Table */ 
 %let pam_recCheckExist=0;
 %let pam_compare_recnt=0;
 %let pam_compare_change = 0;
 %let infileref=WORK.CHANGE; 

 %pam_compare_record_check(input_file=&infileref);
  
 %if %eval(&pam_recCheckExist) gt 0 and &pam_compare_recnt > 0 %then %let pam_compare_change = 1;
                                                               %else %let pam_compare_change = 0;  

%if &pam_compare_change = 0 %then %do; 
  %put NOTE:  No CHANGES identified so no records will be changeg and current_num will be updated at the end ;
%end; 
                                    
%if &pam_compare_change ne 0 %then %do; /* start loop for etl changes found */ 

 /* Update Changed Rows in Target Table */ 
 %put NOTE:  Updating Changed Records ; 
  
    proc sql;
        update &compareds as a
        set 
        %do i=1 %to %eval(&TARGETCNT);
          &&TARGET_FIELD&i = (select &&TARGET_FIELD&i from work.change as b
              where 
             %do x=1 %to &buswords;
                    a.&&bkp&x = b.&&bkp&x
                %if &x ne &buswords %then %do;
                    and
                %end;
                %else %do;
                    )
                %end; 
             %end;
                  
          %if &i lt &TARGETCNT %then %do;
               ,
          %end;
        %end;
      where exists (select * from work.change as b
          where 
         %do x=1 %to &buswords;
              a.&&bkp&x = b.&&bkp&x
             %if &x ne &buswords %then %do;
                    and
                %end;
                %else %do;
                    )
                %end;
         %end;;
     quit; 

 %end;  /* stop loop for change table */   
 
/* ------------------------------------------------------------------------------------------ */  
/* --- Perform the New Append to the Target Table (compareds) from Compare or Change      --- */
/* ------------------------------------------------------------------------------------------ */ 
 /* Record Check on WORK.NEW Table */ 
  %let pam_recCheckExist=0;
  %let pam_compare_recnt=0;
  %let pam_compare_new_change = 0;
  %let pam_compare_new=0; 
  %let pam_compare_recnt=0; 
  %let infileref=WORK.NEW; 
  %pam_compare_record_check(input_file=&infileref);
  
  %if %eval(&pam_recCheckExist) gt 0 and &pam_compare_recnt > 0 %then %let pam_compare_new = 1;
                                                                %else %let pam_compare_new = 0; 
                               
 %if %eval(&pam_compare_new) = 0  %then %do;
    %put NOTE:  No NEW records identified - none appended; 
 %end;
 %else %if &pam_compare_new = 1 %then %do; 
 
 /*---- get the maximum retained key value  ----*/ 
 %let pam_prim_maxkey = -1;
 proc sql noprint;
    select kcompress(put(max(&rkskval), best32.))
       into :pam_prim_maxkey
    from &compareds;
 quit;
 
 %if (&pam_prim_maxkey eq .) %then %let pam_prim_maxkey = 0;
 
 %put pam_prim_maxkey=&pam_prim_maxkey; 
 
/*---- get the last update_dttm from the new to update the load_status table ----*/ 
%pam_determine_latest_loaddttm(libref=WORK,tablename=NEW);


 /* get the single record for the primary key and the first business key used for the new record lookup */ 
 /* update the lookup table with what will be the new _RK or _SK */ 
 proc sort data=work.target_lookup_tbl nodupkey;
   by &bkp1 &rkskval;
 run;  
 

  proc sort data=WORK.NEW(keep=&rkskval &bkp1) out=work.newtest nodupkey;
  by &bkp1 ;
  run; 
  
  data work.new_target_lookup_tbl;
     merge work.newtest(in=a) work.target_lookup_tbl(in=b);
     by &bkp1;
     if _n_ = 1 then do;
        newmaxkey=&pam_prim_maxkey;
     end;
     retain newmaxkey; 
     if a and b then output;
     else if b then output;
     else if a and not b then do;
         newmaxkey=sum(newmaxkey, 1); 
         &rkskval=newmaxkey;
         output; 
     end;
     drop newmaxkey;
  run;
   
  proc sort data=work.new_target_lookup_tbl;
    by &bkp1;
  run; 
  
  proc sort data=WORK.NEW(drop=source_digest buskey_digest compare_digest rc);
    by &bus_prim_key ; 
  run;
  

 data work.newck (keep=&comp_col_list 
                    %do i=1 %to &buswords;
                        &&bkp&i
                    %end;
                    %do x=1%to &pkwords;
                        &&pk&x
                    %end; 
                    CURRENT_NUM PROCESS_DTTM);; 
    
     length &bkp1 &bkp1_type &bkp1_length &RKSKVAL 8 PROCESS_DTTM MISSING 8;  
     retain missing 0 current_num 0 process_dttm 0; 
     if _N_ eq 1 then do;  
       /* build has h0 from lookup table */ 
         nlobs = .;
         dsid = open("WORK.NEW_TARGET_LOOKUP_TBL");
         if (dsid > 0) then do;
             if (attrc(dsid, 'MTYPE') = 'DATA') then
                nlobs = attrn(dsid,'NLOBS');
             else nlobs = -1;
             dsid=close(dsid);
             if (nlobs ne 0) then do;
                if (nlobs > 0) then exponent = ceil(log2(nlobs));
                else exponent = 8;
                                                                                                                            
           declare hash h0(dataset: 'WORK.NEW_TARGET_LOOKUP_TBL', hashexp: exponent);                                                               
              h0.defineKey("&bkp1");                                                                                         
                h0.defineData("&RKSKVAL");                                                                      
           h0.defineDone(); 
           if (nlobs = -1) then do;
              if (h0.Num_items < 1) then do;
                 put "NOTE:  Lookup table is empty - no existing or added items from new";
                 /* abort; */ 
              end;
           end;
        end;
        else do;
           put "NOTE:  Lookup table is empty - no exisitng or added items from new";
            /* abort; */ 
        end;
        end; 
        else do;
          put 'NOTE: Lookup table does not exist or cannot be opened WORK.NEW_TARGET_LKUP_TBL';
         /* abort; */ 
        end; 
        call missing (&bkp1, &RKSKVAL);
      end; 
          
      length errorck 8. ;
      retain errorck; 
      set work.new end=eof;
      /* Check for key */ 
      rc0=h0.find(key: &bkp1);
      /* examine success of lookup */ 
      if (rc0=0) then do;
          output work.newck;
      end;
      else do; 
        errorck + 1; 
        if (rc0 ne 0) then do;
           call missing (&RKSKVAL);
           put missing &bkp1; 
           output work.newck;
        end; 
      end; 
      if eof and errorck > 0 then do; 
         put "Records still missing their primary key value =" errorck; 
      end; 
    run;
    proc sort data=work.newck;
      by &bus_prim_key ;  
    run; 

    data work.newfinal;
      set work.newck;
      by &bus_prim_key ;  
      /* process_dttm picked up from the system date and time run in the pam_prebuild pam_rundtime macro.*/  
      PROCESS_DTTM = &rundtime;
    run; 


  /* to be updated by SPDS or other loader */ 
  /* Compareds is the mart table being appended with the records from the warehouse that are identified as new */ 
    
  proc append base=&compareds
              data=work.newfinal force nowarn;
  run; 
  
   
  %end; /* stop loop for new table */ 
  %else %do;   /* set the last_load_dttm to null - needed for postbuild process */ 
     %let last_dttm = .;
  %end; 
  
   /* --- update the valid_to_dttm and current_num --- */
   /* --- columns in the dimension table as needed --- */

   /* get the indexes (if any) for the dimension table */
   /* so that they can be re-created when finished     */
   /* updating the table                               */
   proc contents data = &compareds 
                 out2 = pam_indexes 
                 noprint;
   run;
   %let pam_indexes=&sysnobs;

   proc sort data = &compareds force;
      by &RKSKVAL descending &pam_fromVar;
   run;

   data &compareds (drop = lag_from);
      set &compareds;
      by &RKSKVAL descending &pam_fromVar;
      lag_from = lag(&pam_fromVar);

      /* the first record is the most recent, so it is current */
      /* and the end is set to the 5999 date if it is invalid  */
      /* or missing                                            */
      if (first.&RKSKVAL) then 
      do; 
         if (&to_dttm eq . or
             &to_dttm lt &pam_fromVar) then 
             &to_dttm = "01Jan5999:00:00:00"dt;
         CURRENT_NUM = 1;
      end;

      /* other records are not current, and the valid_to_dttm    */
      /* &to_dttm represents the valid_to_dttm field             */ 
      /* should be set to the appropriate value:                 */
      /* - it should not be more than the later record's from    */
      /* - it should not be missing                              */
      /* - it should not be less than the current record's from  */
      /* in all of these cases it will be set to 1 second before */
      /* the later record's from date.                           */
      /* If the valid_to_dttm doesn't meet any of these criteria */
      /* then it will be left alone.                             */
      else
      do; 
         if (&to_dttm ge lag_from or
             &to_dttm eq . or
             &to_dttm lt &pam_fromVar) then 
             &to_dttm  = lag_from - 1;
         CURRENT_NUM = 0;
      end;
   run;

   /* put the data back in normal sort mode */
   proc sort data = &compareds;
      by &RKSKVAL &pam_fromVar;
   run;

   /* if there are indexes, then re-create them */
   %if (&pam_indexes gt 0) %then 
   %do; 
      data _null_;
         set pam_indexes end=eof;
         if (_n_ eq 1) then 
         do;
            call execute("proc datasets lib = &comparelib nolist nowarn;");
            call execute("   modify &comparenolib;");
         end;
         call execute(recreate);
         if (eof) then 
            call execute("quit;");
      run;
   %end;

   /* cleanup */
   proc datasets lib = work nolist nowarn;
      delete indexes;
   quit;

%End; /* no source records found for comparison - pam_compare_source section */ 
%End; /* Check for the _RK or _SK for target compare table - required */ 
 
   %if (&debug ne Y and &debug ne 1) %then 
   %do; 
      proc datasets lib = work nolist nowarn memtype = (data view);                                                                                                                                                                   
         delete missing_records nochange pam_compare_digest dict temp pam_master_compare_digest new_target_lookup_tbl                                                                                                                 
                new newck change missing_records newfinal newtest target_lookup_tbl compare_col_list load_ck sourcenolib;                                                                                                                                                   
      quit;
   %end;
  
%end; /* check for parameters  */ 
               
%mend pam_compare;
