

%macro pam_filter_summarize_tag_data;
  /* wrapper around the PAM_DATAREDUCE macro */
  
  
  
  %global P_MAX_OBSERVATIONS g_indsnsize x_varname y_varname;
  %if %nrquote(&x_varname) = %str() %then %let x_varname=MEASURE_DTTM;
  %if %nrquote(&y_varname) = %str() %then %let y_varname=MEASURE;
  
  
  %pam_datareduce(outdsn = &p_outputdataset,
                  indsn = &p_inputdataset,
                  xvar=&x_varname,
                  yvar=&y_varname,
                  maxobs=&P_MAX_OBSERVATIONS,
                  indsnsizevar=g_indsnsize);

%mend;
/****************************************************************************/
/* PAM _DATAREDUCE  
/* Creates a copy of an input data set but reduces the amount of 
/*   observations.  The reduction is based on defining a set of buckets
/*   for a specified X variable in the input data set.  The observation with
/*   the high and low value of a specified Y variable for each bucket are  
/*   placed into the output data set.
/*
/* PARAMETERS
/*    INDSN=    Required input data set  
/*    OUTDSN=   Required output data set
/*    XVAR=     Required variable name whose values will determine the 
/*                 buckets to bucket the data.  The high and low values
/*                 for this variable are subtracted to get a range.  The
/*                 range is divided by the number of buckets to get the size
/*                 of each bucket.  The number of buckets is half of the 
/*                 MAXOBS= value.  It is half because ultimately no more than
/*                 two observations will be placed into each bucket.
/*    YVAR=    Required variable whose high and low values will be saved 
/*                 for each bucket.
/*    MAXOBS=  Maximum number of observations in the output data set.  
/*                 This number also determines the number of buckets
/*                 which will always be half of MAXOBS because no more than 
/*                 two observations will be stored in each bucket.
/*                 This parameter defaults to 10,000.
/*    INDSNSIZEVAR = 
/*              Optional name of a macro variable whose value will be 
/*                 set to the number of observations in the INDSN data set.
/*                 If this variable does not exist, it will be created
/*                 globally.
/**************************************************************************/
%macro pam_datareduce(indsn=, outdsn=, xvar=, yvar=, xvarbegin=,  
                      maxobs=, indsnsizevar=g_indsnsize);
                      
 
%local range bucketsize defaultmaxobs;
%let defaultmaxobs=10000;

/*** parameter validation ***/

/* INDSN */
%if %sysfunc(exist(&indsn)) = 0 %then %do;
   %put ERROR: &sysmacroname must have a valid value for parameter INDSN;
   %goto finish;
%end;

/* OUTDSN */
%if %nrquote(&outdsn) = %str() %then %do;
   %put ERROR: &sysmacroname must have a valid value for parameter OUTDSN;
   %goto finish;
%end;

/* XVAR */
%local xvarnum dsid;
%let dsid = %sysfunc(open(&indsn));
%let xvarnum = %sysfunc(varnum(&dsid, &xvar));
%let dsid = %sysfunc(close(&dsid));
%if &xvarnum <= 0 %then %do;
  %put ERROR: &sysmacroname must have a valid value for parameter XVAR;
  %goto finish;
%end;

/* YVAR */
%local yvarnum;
%let dsid = %sysfunc(open(&indsn));
%let yvarnum = %sysfunc(varnum(&dsid, &yvar));
%let dsid = %sysfunc(close(&dsid));
%if &yvarnum <= 0 %then %do;
  %put ERROR: &sysmacroname must have a valid value for parameter YVAR;
  %goto finish;
%end;

/* MAXOBS */
%if %nrquote(&maxobs) = %str() %then %do;
  %put NOTE: &sysmacroname is using the default value of &defaultmaxobs for the MAXOBS parameter;
  %let maxobs = &defaultmaxobs;
%end;
 
%if %nrquote(&indsnsizevar) ne %str() %then %do;
  %if %sysfunc(nvalid(&indsnsizevar)) = 0 %then %do;
     %put ERROR: If supplied, the optional parameter INDSNSIZEVAR must contain a valid macro variable name.  The supplied value: &&indsnsizevar is not valid.;
     %goto finish;
  %end;
%end;

/*** end of parameter validation ***/

/* get the variables and other DS attributes*/
%local dsid nvars i varname varnames nobs;
%let dsid = %sysfunc(open(&indsn));
%let nvars = %sysfunc(attrn(&dsid, NVARS));
%let nobs = %sysfunc(attrn(&dsid, NOBS)); 

/* if supplied, make sure the name is defined as a global macro variable and set its value to */
/* the number of observations in the input data set */
%if %nrquote(&indsnsizevar) ne %str( ) %then %do;
   %global &indsnsizevar;
   %let &indsnsizevar=&nobs;
%end; 

%if &nobs = 0 %then %do;
   %put WARNING: &sysmacroname:  Input data set &indsn has zero observations.;
%end;
%do i=1 %to &nvars;
   %local varname&i;
   %let varname&i = %sysfunc(varname(&dsid, &i));
%end; 
%let dsid=%sysfunc(close(&dsid)); 
 
/* get the first and last observation value for &XVAR */
%local xvarbegin xvarend rc;
%if &nobs > 0 %then %do;
   %let dsid = %sysfunc(open(&indsn));
   %let rc = %sysfunc(fetch(&dsid));
   %let xvarbegin = %sysfunc(getvarn(&dsid, &xvarnum)); 
   %let rc = %sysfunc(fetchobs(&dsid, &nobs)); 
   %let xvarend = %sysfunc(getvarn(&dsid, &xvarnum));
   %let dsid = %sysfunc(close(&dsid)); 
%end;
 

/* number of buckets is half the number of the MAXOBS */
%local buckets range bucketsize;
%let buckets = %eval(&maxobs / 2);
%if &buckets < 1 %then %let buckets = 1; 
%if %nrquote(&xvarbegin) ne %str() and %nrquote(&xvarend) ne %str() %then 
    %let range = %eval(&xvarend - &xvarbegin);
%else %let range=0;
%let bucketsize = %eval( &range / &buckets );

%put NOTE: &sysmacroname: buckets=&buckets range=&range bucketsize=&bucketsize;

%if &nobs > &maxobs %then %do;
data &outdsn;

   retain minYvalue xValueForMinYValue 999999999999999999 multiple 0 maxYvalue xValueForMaxYValue maxobs minobs threshold beginXvar;
   retain
   %do i = 1 %to &nvars;
      min_&&varname&i max_&&varname&i
   %end;
   ;
   set &indsn ; 

   /* initialize the begin date and initial threshold */
   if _n_=1 then do;
      beginXvar = &xvar; 
      threshold = beginXvar + (multiple * &bucketsize); 
   end;

   /* Keep track of the maximum YVAR value for this bucket and all of */
   /*   the values for this observation */
   if &yvar > maxYvalue then do;
      maxYvalue=&yvar;
      xValueForMaxYValue=&xvar;
      maxobs=_n_;
      %do i = 1 %to &nvars;
         max_&&varname&i = &&varname&i;
      %end;
   end;

   /* Keep track of the minimum YVAR value for this bucket and all of */
   /*   the values for this observation */
   if &yvar < minYvalue then do;
      minYvalue=&yvar;
      xValueForMinYValue = &xvar;
      minobs=_n_;
      %do i = 1 %to &nvars;
         min_&&varname&i = &&varname&i;
      %end;
   end;


   /* if the XVAR value has crossed the next threshold then output the */
   /*   running maxYvalue and minYvalue observation values, then reset everything  */
   /*   and increase the multiple to the next threshold that is greater */
   /*   than the current XVAR value.   */
   if &xvar >= threshold then do;

      /* to guarantee the output data is ordered by the XVAR values, the XVAR value */
      /* for the maximum and minimum YVAR value are compared.  Whichever is lowest */
      /* determines which point is output first */ 
      if xValueForMaxYValue lt xValueForMinYValue then do;
         if maxobs ne . then do;
            %do i = 1 %to &nvars;
               &&varname&i = max_&&varname&i;
            %end;
            origobs = maxobs;   
            output;
         end;
         if minobs ne . and minobs ne maxobs then do;
            %do i = 1 %to &nvars;
               &&varname&i = min_&&varname&i;
            %end;
            origobs = minobs;  
            output;
         end;
      end;
      else do;
         if minobs ne . and minobs ne maxobs then do;
            %do i = 1 %to &nvars;
               &&varname&i = min_&&varname&i;
            %end;
            origobs = minobs;   
            output;
         end;
         if maxobs ne . then do;
            %do i = 1 %to &nvars;
               &&varname&i = max_&&varname&i;
            %end;
            origobs = maxobs;   
            output;
         end;
      end;

      /* skip to the next threshold that is greater than the current xvar value.  This may involve skipping */
      /* multiple thresholds.  Then reset the running max and min values */
      do until (&xvar < threshold);
         multiple + 1;
         threshold = beginXvar + (multiple * &bucketsize);
      end;
      maxYvalue=.;
      maxobs=.;
      xValueForMaxYValue=.;
      xValueForMinYValue=.;
      minYvalue=99999999999999999999999; *<== this has to be set to a very large number so that the next value will always be less;
      minobs=.;
   end;

   drop minYvalue xValueForMinYValue maxYvalue xValueForMaxYValue maxobs minobs beginXvar origobs multiple;
   drop
   %do i = 1 %to &nvars;
      min_&&varname&i max_&&varname&i
   %end;
   ;
   run;
%end;
%else %do;
   %put NOTE: &sysmacroname:  Input data set size of &nobs is smaller than the MAXOBS value of &maxobs..  No data reduction will occur.;
/**** we need to see if we can just reuse the input data set rather than copying it */
   data &outdsn;
      set &indsn;
   run;
%end;
%finish:
%mend;
